/*
 * Mach Operating System
 * Copyright (c) 1992, 1991 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */
/*
 * Amiga kernel console code
 *
 * Contributors:
 *      Bryan Ford
 */

#include <mach/boolean.h>
#include <machine/machparam.h>
#include <machine/kd.h>
#include <machine/uart.h>


static boolean_t inited;


/* Called from init.c */
void
cninit(void)
{
  kdinit();
  utinit();
  inited = 1;
}


/* Send a character from the kernel to both the screen and the serial port */
void
cnputc(u_char c)
{
  if(inited)
    {
      int s;

      if(c == '\n')
        cnputc('\r');
      s = splhigh();
      ut_cnputc(c);
      kd_cnputc(c);
      splx (s);
    }
}


/* Get a key from either the keyboard or the serial port */
int
cngetc(void)
{
  int s, c, c_o;

  if(!inited)
    return(-1);

  do
    {
      s = splhigh();
      if ((c = ut_cnpollc()) < 0)
        c = kd_cnpollc();
      splx (s);
    }
  while (c < 0);

  return(c);
}

